import json

import click

import ghchain
from ghchain.utils import run_command


def handle_land(branch):
    """
    Merge the specified branch into the configured base branch.
    This will close the corresponding PRs and delete the branch if configured to do so.
    """

    if (
        branch not in ghchain.repo.branches
        or ghchain.repo.branches[branch].commit
        != ghchain.repo.remotes.origin.refs[branch].commit
    ):
        # if the branch is not up to date, pull the latest changes
        if click.confirm(
            f"The local branch '{branch}' is out of date with the remote. Do you want to update it?"
        ):
            try:
                run_command(
                    ["git", "fetch", "origin", f"{branch}:{branch}"], check=True
                )
            except Exception as e:
                ghchain.logger.error(
                    f"Failed to pull the latest changes from the remote: {e}"
                )
                return
        else:
            ghchain.logger.error(
                f"Branch '{branch}' is not up to date with the remote. Aborting."
            )
            return

    # update the base branch
    run_command(
        [
            "git",
            "branch",
            "-f",
            ghchain.config.base_branch.replace("origin/", ""),
            branch,
        ],
        check=True,
    )
    run_command(
        ["git", "push", "origin", ghchain.config.base_branch.replace("origin/", "")],
        check=True,
    )

    if ghchain.config.delete_branch_after_merge:
        # List all open PRs targeting the branch to be deleted
        prs = json.loads(
            run_command(
                [
                    "gh",
                    "pr",
                    "list",
                    "--json",
                    "number",
                    "--state",
                    "open",
                    "--base",
                    branch,
                ],
                check=True,
            ).stdout
        )

        # Change the base branch of all those PRs to target the configured base branch
        for pr in prs:
            run_command(
                [
                    "gh",
                    "pr",
                    "edit",
                    str(pr["number"]),
                    "--base",
                    ghchain.config.base_branch.replace("origin/", ""),
                ],
                # Don't check here because somethimes github updates the PRs faster than the CLI
                check=False,
            )

        # Delete the remote branch, don't check since the branch might already be deleted
        run_command(["git", "push", "origin", "--delete", branch], check=False)
        # Delete the local branch
        run_command(["git", "branch", "-D", branch], check=True)



import re
from typing import Optional

import pandas as pd
from pydantic import BaseModel

import ghchain
from ghchain.git_utils import get_issue_url
from ghchain.pull_request import PR
from ghchain.status import (
    PrStatus,
    StatusCheck,
    WorkflowStatus,
    mergeable_status_to_ansi,
    review_decision_to_ansi,
)


def status_to_emoji(status: str):
    status_to_emoji_mapping = {
        "success": "✅",
        "failure": "❌",
        "neutral": "⚪",
        "cancelled": "🛑",
        "timed_out": "⏰",
        "in_progress": "🔄",
        "completed": "✅",
    }

    if status in status_to_emoji_mapping:
        return status_to_emoji_mapping[status]
    else:
        return status


def status_check_to_note(status_checks: dict[str, StatusCheck]):
    status_checks_df = pd.DataFrame(
        [
            {
                k: status_to_emoji(v.lower())
                for k, v in status_check.__dict__.items()
                if k in ["name", "conclusion", "status"]
            }
            for status_check in status_checks.values()
        ]
    ).set_index("name")

    return "\n\n[[status_checks]]\n\n" + status_checks_df.to_markdown()


def workflow_statuses_to_note(workflow_statuses: list[WorkflowStatus]):
    workflow_statuses_df = pd.DataFrame(
        [
            {
                k: status_to_emoji(v)
                for k, v in ws.to_dict().items()
                if k in {"status", "conclusion", "name"}
            }
            for ws in workflow_statuses
        ]
    ).set_index("name")
    return "\n\n[[workflow_statuses]]\n\n" + workflow_statuses_df.to_markdown()


def get_commit_notes(
    pr_url: str | None = None,
    pr_status: PrStatus | None = None,
    issue_url: str | None = None,
    workflow_statuses: list[WorkflowStatus] | None = None,
):
    if issue_url:
        return f"[ghchain]\nissue = {issue_url}\n"
    if not pr_status and not pr_url and not workflow_statuses:
        return ""

    notes_str = f"[ghchain]\npr url = {pr_url}\n"
    if pr_status:
        notes_str += (
            f"Review Decision = {review_decision_to_ansi(pr_status.review_decision)}\n"
        )
        notes_str += f"Mergable = {mergeable_status_to_ansi(pr_status.is_mergeable)}\n"
        notes_str += "\n".join(
            [
                f"{key} = {value}"
                for key, value in pr_status.to_dict().items()
                if key in ["is_draft", "title"]
            ]
        )

    # convert workflow_statuses to a markdonw table
    if workflow_statuses:
        notes_str += workflow_statuses_to_note(workflow_statuses)
    if pr_status and pr_status.status_checks:
        notes_str += status_check_to_note(pr_status.status_checks)

    return notes_str


class Commit(BaseModel):
    """
    Commit object with sha, message, branch, pr_url, notes, and pr_status.
    It is expected that every non-fixup commit has its corresponding branch.
    """

    sha: str
    message: str
    branch: str | None = None
    # if the commit is linked to an issue:
    issue_url: str | None = None
    pull_request: PR | None = None
    notes: Optional[str] = None
    workflow_statuses: Optional[list[WorkflowStatus]] = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)

        # extract the linked issue id from the commit message
        self.issue_url = self.extract_issue_url()

        self.workflow_statuses = []
        for workflow in ghchain.config.workflows:
            workflow_status = WorkflowStatus.from_commit_id(
                workflow, commit_id=data["sha"]
            )
            if workflow_status:
                self.workflow_statuses.append(workflow_status)

        # Fetch and parse commit notes
        self.notes = get_commit_notes(
            pr_url=self.pull_request.pr_url if self.pull_request else None,
            pr_status=self.pull_request.pr_status if self.pull_request else None,
            issue_url=self.issue_url,
            workflow_statuses=self.workflow_statuses,
        )

        self.update_notes()

    def extract_issue_id(self) -> Optional[int]:
        """
        Extract the issue ID from the commit message if it contains the GitHub issue id
        in the format: [issue tags] commit header (#issue id)
        """
        pattern = re.compile(ghchain.config.issue_pattern)
        match = pattern.search(self.message)

        if match:
            return int(match.group(1))
        return None

    def extract_issue_url(self) -> Optional[str]:
        """
        Extract the issue URL from the commit message.
        """
        issue_id = self.extract_issue_id()
        if issue_id:
            return get_issue_url(issue_id)
        return None

    def update_notes(self):
        # Write the updated notes back to the git notes

        if self.notes:
            ghchain.repo.git.notes("add", "-f", "-m", self.notes, self.sha)

    @property
    def pr_id(self) -> Optional[int]:
        if not self.pr_url:
            return None
        return int(self.pr_url.rstrip("/").split("/")[-1])

    @property
    def is_fixup(self) -> bool:
        return self.message.startswith("fixup!") or self.message.startswith("squash!")



from typing import List, Optional

import click
from loguru import logger
from pydantic import BaseModel

import ghchain
from ghchain.commit import Commit
from ghchain.git_utils import (
    create_branch_name,
    get_all_branches,
    get_current_branch,
    git_push,
)
from ghchain.github_utils import (
    create_branch_from_issue,
    get_next_gh_id,
)
from ghchain.pull_request import (
    PR,
    get_open_prs,
    run_tests_on_branch,
    update_pr_descriptions,
)
from ghchain.utils import run_command


class Stack(BaseModel):
    commits: List[Commit]
    dev_branch: str
    base_branch: str

    @classmethod
    def create(cls, base_branch: Optional[str] = None) -> "Stack":
        """
        Create a Stack object with commits from the current branch which are not in the base branch.

        Args:
            base_branch: The base branch to compare the commits against. Defaults to the base branch in the config.
            dev_branch: The development branch to create the stack from. Defaults to the current branch.

        """

        sha_to_pull_request_mapping: dict[str, PR] = {
            pr.commits[-1]: pr for pr in get_open_prs()
        }

        if not base_branch:
            base_branch = ghchain.config.base_branch
        dev_branch = get_current_branch()

        logger.debug(f"Creating stack with dev branch: {dev_branch}")

        # Fetch the commits that are in dev_branch but not in base_branch
        commits_diff = ghchain.repo.iter_commits(f"{base_branch}..{dev_branch}")

        commits = []
        # Reverse log output to start from the bottom of the stack
        for commit in reversed(list(commits_diff)):
            sha = commit.hexsha
            message = commit.message.strip()
            is_fixup = message.startswith("fixup!")

            # Find branches that point to this commit (branches where this commit is the HEAD)
            pointing_branches = [
                ref.name.split("/")[-1]
                for ref in ghchain.repo.branches
                if ref.commit == commit
            ]
            pointing_branches = set(pointing_branches) - {
                dev_branch
            }  # Exclude the dev branch

            if len(pointing_branches) > 1:
                error_message = f"Commit {sha} has multiple branches: {pointing_branches}. This is not supported."
                logger.error(error_message)
                raise ValueError(error_message)

            branch = pointing_branches.pop() if pointing_branches else None
            commit_obj = Commit(
                sha=sha,
                branch=branch,
                message=message,
                is_fixup=is_fixup,
                pull_request=sha_to_pull_request_mapping.get(sha),
            )
            commits.append(commit_obj)

            if is_fixup:
                # Find the previous commit that is not a fixup and assign it the same branch
                target_commit = next(
                    (
                        commit
                        for commit in commits[::-1]
                        if not commit.is_fixup and not commit.branch
                    ),
                    None,
                )
                if target_commit:
                    target_commit.branch = commit_obj.branch

        return cls(commits=commits, dev_branch=dev_branch, base_branch=base_branch)

    @property
    def commit2idx(self):
        return {commit.sha: i for i, commit in enumerate(self.commits)}

    @property
    def branch_ids(self) -> set[int]:
        return {
            int(commit.branch.split("-")[-1])
            for commit in self.commits
            if commit and commit.branch
            if commit.branch.split("-")[-1].isnumeric()
        }

    @property
    def branches(self) -> list[str]:
        """
        Return the branches in the stack, sorted by order in the stack.
        """
        return [commit.branch for commit in self.commits if commit.branch]

    def process_commit(
        self,
        commit: Commit,
        create_pr: bool = False,
        draft: bool = False,
        with_tests: bool = False,
    ):
        """
        Create a branch for each commit in the stack that doesn't already have one.
        If create_pr is True, create a PR for each branch.
        """
        pr_created = False
        if not commit.branch:
            if commit.issue_url:
                logger.info(f"Creating branch from issue {commit.issue_url}")
                # set a temporary branch name as it will be created later
                branch_name = "{will be created from issue}"

            else:
                branch_id = max([get_next_gh_id(), *[id + 1 for id in self.branch_ids]])
                branch_name = create_branch_name(
                    ghchain.config.branch_name_template, branch_id
                )
            click.confirm(
                f"Create branch {branch_name} for commit {commit.sha}?\n{commit.message}\n",
                abort=True,
                default=True,
            )
            logger.info(f"Creating branch {branch_name} for commit {commit.sha}")
            if commit.issue_url:
                # create the branch from the gh issue
                branch_name = create_branch_from_issue(
                    issue_id=int(commit.issue_url.split("/")[-1]),
                    base_commit=commit.sha,
                )
            else:
                ghchain.repo.git.branch(branch_name, commit.sha)
            commit.branch = branch_name

            # Push the branch to the remote
            git_push(branch_name)

        else:
            branch_name = commit.branch

        # If the commit is not a fixup and it does not have a PR, create one
        if not commit.is_fixup and create_pr and not commit.pull_request:
            commit.pull_request = PR.create_pull_request(
                base_branch=ghchain.config.base_branch.replace("origin/", "")
                if not self.commit2idx[commit.sha]
                else self.commits[self.commit2idx[commit.sha] - 1].branch,
                head_branch=branch_name,
                title=commit.message.split("\n")[0],
                body=commit.message,
                draft=draft,
                commit_sha=commit.sha,
            )
            pr_created = True

        if create_pr:
            update_pr_descriptions(
                pr_stack=[
                    c.pull_request
                    for c in self.commits[: self.commit2idx[commit.sha] + 1]
                    if c.pull_request
                ]
            )

        if with_tests:
            run_tests_on_branch(branch_name, commit.pull_request)

        return pr_created

    def publish(self):
        """
        Collect all branches in the stack that need to be pushed to the remote,
        ask the user for confirmation, and then push them all with force-with-lease.
        """
        branches_to_push = []

        # Collect branches that need to be pushed
        for commit in self.commits:
            if not commit.branch:
                logger.trace(
                    f"Commit {commit.sha} does not have an associated branch. Skipping."
                )
                continue

            branch_name = commit.branch
            remote_branch = f"origin/{branch_name}"

            try:
                local_sha = ghchain.repo.git.rev_parse(branch_name)
                remote_sha = ghchain.repo.git.rev_parse(remote_branch)

                if local_sha != remote_sha:
                    branches_to_push.append(branch_name)
            except Exception:
                # Remote branch does not exist
                branches_to_push.append(branch_name)

        if not branches_to_push:
            logger.info("All branches are up-to-date with the remote. Nothing to push.")
            return

        # Confirm with the user
        click.confirm(
            f"The following branches will be pushed with --force-with-lease:\n{', '.join(branches_to_push)}\n"
            "Do you want to proceed?",
            abort=True,
        )

        # Push all branches with --force-with-lease
        try:
            ghchain.repo.git.push("--force-with-lease", "origin", *branches_to_push)
            logger.info(
                f"Successfully pushed branches: {', '.join(branches_to_push)} with --force-with-lease."
            )
        except Exception as push_error:
            logger.error(
                f"Failed to push branches: {', '.join(branches_to_push)}: {push_error}"
            )

        logger.info("Publishing complete.")


def find_branch_of_rebased_commit(commit: Commit) -> str:
    """
    When a commit is rebased, the commit hash changes. This function checks if there is any branch
    that has a commit with the message as top commit. If there is, it returns the branch name.
    """
    # Get the commit message
    commit_message = commit.message

    # List all branches
    result = get_all_branches()
    branches = result.stdout.splitlines()

    # Check the top commit message of each branch
    for branch in branches:
        branch = branch.strip().replace(
            "* ", ""
        )  # Remove leading '*' for the current branch
        result = run_command(
            ["git", "log", "-1", "--pretty=%B", branch],
            check=True,
        )
        top_commit_message = result.stdout.strip()
        if top_commit_message == commit_message:
            logger.debug(f"Found branch {branch} with commit message {commit_message}")
            return branch

    return None



import sys
from unittest.mock import MagicMock

import pytest
from git import Repo


def is_debugging():
    """
    Check if pytest is run from a debugger.
    """
    return "debugpy" in sys.modules


# enable_stop_on_exceptions if the debugger is running during a test
# taken from https://stackoverflow.com/questions/62419998/how-can-i-get-pytest-to-not-catch-exceptions/62563106#62563106
if is_debugging():

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


@pytest.fixture
def mock_repo(monkeypatch):
    repo = MagicMock(spec=Repo)
    monkeypatch.setattr("ghchain.repo", repo)
    return repo



import subprocess
from pathlib import Path
from typing import List

import tomllib
from loguru import logger
from pydantic import BaseModel, ConfigDict, Field, field_validator


class Config(BaseModel):
    model_config = ConfigDict(frozen=True)

    workflows: List[str] = Field(default_factory=list)
    git_username: str = Field(default="")
    base_branch: str = Field(default="origin/main")
    branch_name_template: str = Field(default="{git_config_author}-{pr_id}")
    delete_branch_after_merge: bool = Field(default=True)

    # regex pattern to detect linked issues in commit messages
    issue_pattern: str = Field(default=r"\(#(\d+)\)")

    # logging
    log_file: Path | None = Field(default=None)
    log_level: str = Field(default="INFO")

    @field_validator("git_username")
    @classmethod
    def set_git_username(cls, v):
        return v or subprocess.getoutput("git config user.name")

    @classmethod
    def from_toml(cls, toml_fn: Path):
        if not toml_fn.exists():
            logger.warning(f"No config file found at {toml_fn}. Using default values.")
            return cls()
        toml_string = toml_fn.read_text()
        toml_dict = tomllib.loads(toml_string)
        return cls(**toml_dict)

    def to_dict(self):
        return self.model_dump()



"""Top-level package for ghchain."""

__author__ = """Hendrik Klug"""
__email__ = "hendrik.klug@gmail.com"
__version__ = "0.2.0"


import sys
from pathlib import Path

from git import Repo
from loguru import logger

from ghchain.config import Config

try:
    repo = Repo(".")
except Exception:
    logger.error("Could not find a git repository in the current directory.")
    sys.exit(1)

config = Config.from_toml(Path(repo.git_dir).parent / ".ghchain.toml")
logger.remove()
logger.add(sys.stderr, level=config.log_level)
if config.log_file:
    logger.add(config.log_file, level=config.log_level)


if (
    toml_fn := Path(repo.git_dir)
    .parent.parent.parent.joinpath(".ghchain.toml")
    .exists()
):
    logger.info(f"Loaded config from {toml_fn}")
logger.debug(f"Config: {config.to_dict()}")



import os
import subprocess
from typing import Optional

from ghchain.config import logger


def run_command(
    command: list[str],
    check=False,
    env: Optional[dict] = None,
    shell=False,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
) -> subprocess.CompletedProcess:
    try:
        cwd = os.getcwd()
        logger.debug(f"Running command: \"{' '.join(command)}\" in cwd: {cwd}")
        result = subprocess.run(
            command,
            stdout=stdout,
            stderr=stderr,
            text=True,
            check=check,
            env=env,
        )
        logger.debug(f"Command completed with return code: {result.returncode}")
        logger.trace(f"Command stdout: {result.stdout}")
        logger.trace(f"Command stderr: {result.stderr}")
        return result
    except subprocess.CalledProcessError as e:
        logger.error(
            f"Command '{' '.join(command)}' failed with error: {str(e)} in cwd: {cwd}"
        )
        logger.error(f"Stdout: {e.stdout}")
        logger.error(f"Stderr: {e.stderr}")
        raise e



import json
import os
import subprocess
import time
from pathlib import Path

import pytest
from click.testing import CliRunner
from git import Repo

from ghchain import cli
from ghchain.config import Config, logger
from ghchain.git_utils import get_all_branches
from ghchain.stack import Stack
from ghchain.utils import run_command


@pytest.fixture(scope="session")
def setup_mytest_repo(tmpdir_factory):
    temp_dir = tmpdir_factory.mktemp("shared_temp_dir")
    os.chdir(temp_dir)

    logger.info(f"Current working directory: {os.getcwd()}")
    subprocess.run(
        ["git", "clone", "https://github.com/HendrikKlug-synthara/mytest.git", "."],
        capture_output=True,
        text=True,
        cwd=os.getcwd(),
        check=True,
    )

    if "GH_TOKEN" in os.environ:
        command = (
            f"git remote set-url origin "
            f"https://x-access-token:{os.environ['GH_TOKEN']}@"
            f"github.com/HendrikKlug-synthara/mytest.git"
        )
        subprocess.run(command, shell=True, check=True)
    return temp_dir


@pytest.fixture
def patch_git_repo(monkeypatch, setup_mytest_repo):
    repo = Repo(setup_mytest_repo)
    monkeypatch.setattr("ghchain.repo", repo)


@pytest.fixture
def patch_config(monkeypatch, setup_mytest_repo):
    """
    Overwrite the .ghchain with a custom configuration for testing.
    """
    mytest_config = """\
 workflows = ["workflow_1", "workflow_2"]
 branch_name_template = "hk-{pr_id}"
 log_level = "TRACE"
 issue_pattern = "\\\\(#(\\\\d+)\\\\)"
 """
    ghchain_config = Path(setup_mytest_repo) / ".ghchain.toml"
    ghchain_config.write_text(mytest_config)
    config = Config.from_toml(ghchain_config)
    monkeypatch.setattr("ghchain.config", config)


@pytest.fixture
def cli_runner(setup_mytest_repo, patch_git_repo, patch_config):
    runner = CliRunner()
    yield runner, setup_mytest_repo


@pytest.fixture(scope="function")
def repo_cleanup():
    """Fixture to clean up the repository before tests."""
    cleanup_repo()
    yield


def commit_fixup(branch):
    run_command(["git", "checkout", branch])
    run_command(["touch", "new_file"])
    run_command(["git", "add", "new_file"])
    run_command(["git", "commit", "-m", "fixup! new file"])
    run_command(["git", "push"])
    run_command(["git", "checkout", "-"])


def cleanup_repo():
    """Delete all current branches and pull requests and create a new branch mydev."""
    run_command(["git", "checkout", "main"])
    run_command(["git", "reset", "--hard", "464397e0bf88c66ea09fd05448ad6847c230c2fe"])
    run_command(["git", "push", "-f"])

    local_branches = set(get_all_branches()) - {"main"}
    remote_branches = set(get_all_branches(remote=True)) - {"main"}

    for branch in local_branches:
        run_command(["git", "branch", "-D", branch])
    for branch in remote_branches:
        run_command(["git", "push", "origin", "--delete", branch])

    prs_json = run_command(
        ["gh", "pr", "list", "--json", "url"],
    ).stdout

    prs = json.loads(prs_json)
    for pr in prs:
        pr_url = pr["url"]
        run_command(
            ["gh", "pr", "close", pr_url, "-d"],
        )

    run_command(["git", "checkout", "main"])
    run_command(["git", "checkout", "-b", "mydev"])


def create_stack():
    """Change the README.md file 4 times and commit each change."""
    for i in range(4):
        with open("README.md", "a") as f:
            f.write(f"line {i}\n")
        run_command(["git", "add", "README.md"])
        run_command(["git", "commit", "-m", f"commit {i}\nThis is a test commit."])


@pytest.mark.order(1)
@pytest.mark.parametrize("create_pr", [True, False])
@pytest.mark.parametrize("draft", [True, False])
@pytest.mark.parametrize("with_tests", [True, False])
def test_process_commits(cli_runner, repo_cleanup, create_pr, draft, with_tests):
    cli_runner, _ = cli_runner

    create_stack()

    args = []
    if create_pr:
        args.append("--create-pr")
    if draft:
        args.append("--draft")
    if with_tests:
        args.append("--with-tests")

    result = cli_runner.invoke(cli.ghchain_cli, args)
    if result.exit_code != 0:
        print(result.output)

    assert (
        result.exit_code == 0
    ), f"Command failed with exit code {result.exit_code}\n{result.output}"

    stack = Stack.create()
    assert len(stack.commits) == 4
    if create_pr or draft:
        for commit in stack.commits:
            assert commit.branch is not None

    # assert that the repo has 4 open pull requests
    if create_pr or draft:
        prs = run_command(["gh", "pr", "list", "--json", "baseRefName"]).stdout
        prs = json.loads(prs)
        assert len(prs) == 4, f"Expected 4 pull requests, got {len(prs)}"

        branches = (["main"] + [commit.branch for commit in stack.commits[:-1]])[::-1]
        for idx, pr in enumerate(prs):
            assert pr["baseRefName"] == branches[idx]


@pytest.mark.order(1)
def test_rebase(cli_runner, repo_cleanup):
    cli_runner, _ = cli_runner

    # Stack is up to date with origin/main
    create_stack()
    result = cli_runner.invoke(cli.ghchain_cli)

    assert result.exit_code == 0

    stack = Stack.create(base_branch="main")
    bottom_branch = stack.branches[0]

    commit_fixup(bottom_branch)
    result = cli_runner.invoke(cli.rebase, [bottom_branch])

    assert result.exit_code == 0

    stack = Stack.create()
    assert len(stack.commits) == 5


@pytest.mark.order(1)
def test_land(cli_runner, repo_cleanup):
    cli_runner, _ = cli_runner

    create_stack()

    result = cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])
    assert result.exit_code == 0

    stack = Stack.create(base_branch="main")
    branch_to_land = stack.branches[0]
    runner = CliRunner()
    result = runner.invoke(cli.land, ["-b", branch_to_land])
    assert result.exit_code == 0

    # assert that the repo has 3 open pull requests
    prs = run_command(["gh", "pr", "list", "--json", "url"]).stdout
    prs = json.loads(prs)
    assert len(prs) == 3, f"Expected 3 pull requests, got {len(prs)}"


@pytest.mark.order(1)
def test_land_local_out_of_date(cli_runner, repo_cleanup):
    cli_runner, _ = cli_runner

    create_stack()

    result = cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])
    assert result.exit_code == 0

    # make the branch to merge out of date with remote
    stack = Stack.create(base_branch="main")
    bottom_branch = stack.branches[0]
    bottom_commit = stack.commits[0].sha
    commit_fixup(bottom_branch)
    # reset the branch to the original commit
    run_command(["git", "checkout", bottom_branch])
    run_command(["git", "reset", "--hard", bottom_commit])
    run_command(["git", "checkout", "-"])

    runner = CliRunner()
    result = runner.invoke(cli.land, ["-b", bottom_branch], input="y\n")
    assert result.exit_code == 0

    # assert that the repo has 3 open pull requests
    prs = run_command(["gh", "pr", "list", "--json", "url"]).stdout
    prs = json.loads(prs)
    assert len(prs) == 3, f"Expected 3 pull requests, got {len(prs)}"


@pytest.mark.order(1)
def test_land_no_local_branch(cli_runner, repo_cleanup):
    cli_runner, _ = cli_runner

    create_stack()

    result = cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])
    assert result.exit_code == 0

    # make the branch to merge out of date with remote
    stack = Stack.create(base_branch="main")
    bottom_branch = stack.branches[0]

    # delete the local branch
    run_command(["git", "branch", "-D", bottom_branch])

    runner = CliRunner()
    result = runner.invoke(cli.land, ["-b", bottom_branch], input="y\n")
    assert result.exit_code == 0

    # assert that the repo has 3 open pull requests
    prs = run_command(["gh", "pr", "list", "--json", "url"]).stdout
    prs = json.loads(prs)
    assert len(prs) == 3, f"Expected 3 pull requests, got {len(prs)}"


@pytest.mark.order(2)
def test_multiple_commits_per_pr(cli_runner, repo_cleanup):
    """
    Test process commits when there are commits added to a pr.
    """
    cli_runner, _ = cli_runner

    # Stack is up to date with origin/main
    create_stack()
    cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])

    stack = Stack.create(base_branch="main")
    bottom_branch = stack.branches[-1]

    run_command(["git", "checkout", bottom_branch])
    run_command(["touch", "new_file"])
    run_command(["git", "add", "new_file"])
    run_command(["git", "commit", "-m", "fixup! new file"])
    run_command(["git", "push"])
    run_command(["git", "checkout", "-"])
    cli_runner.invoke(cli.rebase, [bottom_branch])

    stack = Stack.create()

    result = cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])
    assert result.exit_code == 0

    # assert that the repo has 4 open pull requests
    prs = run_command(["gh", "pr", "list", "--json", "url"]).stdout
    prs = json.loads(prs)
    assert len(prs) == 4, f"Expected 4 pull requests, got {len(prs)}"


@pytest.mark.order(2)
def test_main_out_of_date(cli_runner, repo_cleanup):
    """
    Test that process commits works also if the local main branch is out of date.
    """
    cli_runner, _ = cli_runner

    # Stack is up to date with origin/main
    create_stack()

    run_command(["git", "checkout", "main"])
    run_command(["git", "reset", "--hard", "HEAD~2"])
    run_command(["git", "checkout", "-"])

    result = cli_runner.invoke(cli.ghchain_cli, ["--create-pr"])

    # assert that the repo has 4 open pull requests
    prs = run_command(["gh", "pr", "list", "--json", "url"]).stdout
    prs = json.loads(prs)
    assert len(prs) == 4, f"Expected 4 pull requests, got {len(prs)}"

    assert result.exit_code == 0


@pytest.mark.order(2)
@pytest.mark.parametrize("with_prs", [True, False])
def test_run_tests(cli_runner, repo_cleanup, with_prs):
    """
    Test the run-tests command.
    Run the workflows for the first commit in the stack.
    """
    cli_runner, _ = cli_runner

    # create stack
    create_stack()
    command_args = ["--create-pr"] if with_prs else []
    result = cli_runner.invoke(cli.ghchain_cli, command_args)
    assert result.exit_code == 0

    stack = Stack.create()
    result = cli_runner.invoke(cli.run_workflows, ["-b", stack.commits[0].branch])

    time.sleep(30)
    stack = Stack.create()

    # verify that the commit notes have been updated
    assert (
        "[[workflow_statuses]]" in stack.commits[0].notes
    ), f"Expected [[workflow_statuses]] in notes, got {stack.commits[0].notes}"

    assert result.exit_code == 0


@pytest.mark.order(1)
def test_linked_issue(cli_runner, repo_cleanup):
    """
    Test that commits with a linked issue in the message are correctly detected
    and the branch is linked to the corresponding GitHub issue.
    """
    cli_runner, _ = cli_runner

    # Step 1: Create a new GitHub issue
    issue_title = "Test issue for linking"
    issue_body = "This issue is created for testing branch linking in ghchain."
    issue_creation_result = run_command(
        [
            "gh",
            "issue",
            "create",
            "--title",
            issue_title,
            "--body",
            issue_body,
        ],
    )
    assert (
        issue_creation_result.returncode == 0
    ), f"Issue creation failed: {issue_creation_result.stderr}"

    # Extract the issue number from the output
    issue_url = issue_creation_result.stdout.strip()
    issue_id = int(issue_url.split("/")[-1])

    # Step 2: Create a commit with a linked issue reference
    run_command(["git", "checkout", "mydev"])
    with open("README.md", "a") as f:
        f.write("This commit is linked to an issue.\n")
    run_command(["git", "add", "README.md"])
    commit_message = f"Add feature linked to issue (#{issue_id})"
    run_command(["git", "commit", "-m", commit_message])

    # Step 3: Run ghchain to process the commit
    result = cli_runner.invoke(cli.ghchain_cli)
    assert (
        result.exit_code == 0
    ), f"Command failed with exit code {result.exit_code}\n{result.output}"

    # Step 4: Check that the commit was processed correctly
    stack = Stack.create()
    assert len(stack.commits) == 1, "Expected 1 commit in the stack"

    commit = stack.commits[0]
    assert (
        commit.issue_url == issue_url
    ), f"Expected issue_id {issue_url}, got {commit.issue_url}"
    assert commit.branch is not None, "Expected branch to be created for the commit"

    # Step 5: Verify that the branch is linked to the GitHub issue
    issue_branches = run_command(
        ["gh", "issue", "develop", "--list", str(issue_id)]
    ).stdout
    assert issue_branches.split("\t")[0] == commit.branch, (
        f"Expected branch {commit.branch} to be linked to issue {issue_id}, "
        f"got {issue_branches}"
    )


if __name__ == "__main__":
    pytest.main(["-v", __file__, "-s", "-x", "-k linked_issue"])



import json
from dataclasses import dataclass
from enum import Enum, StrEnum
from typing import Any, Dict, Optional

import ghchain
from ghchain.github_utils import get_pr_id_for_branch
from ghchain.utils import run_command


@dataclass
class WorkflowStatus:
    status: str
    conclusion: str
    name: str
    branch: str
    commit: str
    event: str
    id: str
    elapsed: str
    timestamp: str
    workflow_yml_fn: str

    @classmethod
    def from_line(cls, line: str, workflow_yml_fn: str):
        status, conclusion, name, branch, commit, event, id, elapsed, timestamp = (
            line.split("\t")
        )
        return cls(
            status=status,
            conclusion=conclusion,
            name=name,
            branch=branch,
            commit=commit,
            event=event,
            id=id,
            elapsed=elapsed,
            timestamp=timestamp,
            workflow_yml_fn=workflow_yml_fn,
        )

    @classmethod
    def create(cls, workflow_yml_fn: str, branchname: str):
        result = run_command(
            [
                "gh",
                "run",
                "list",
                "--workflow",
                f"{workflow_yml_fn}.yml",
                "-b",
                branchname,
            ],
            check=True,
        )
        lines = result.stdout.splitlines()
        if lines:
            line = result.stdout.splitlines()[0]
            return WorkflowStatus.from_line(line, workflow_yml_fn)
        return None

    @classmethod
    def from_commit_id(cls, workflow_yml_fn: str, commit_id: str):
        result = run_command(
            [
                "gh",
                "run",
                "list",
                "--commit",
                commit_id,
                "--workflow",
                f"{workflow_yml_fn}.yml",
            ],
            check=True,
        )
        lines = result.stdout.splitlines()
        if lines:
            line = result.stdout.splitlines()[0]
            return WorkflowStatus.from_line(line, "")
        return None

    def to_dict(self) -> Dict[str, any]:
        return self.__dict__


class StatusCheckConclusion(StrEnum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class StatusCheckStatus(StrEnum):
    COMPLETED = "COMPLETED"


@dataclass
class StatusCheck:
    completedAt: str
    conclusion: StatusCheckConclusion
    detailsUrl: str
    name: str
    startedAt: str
    status: StatusCheckStatus
    workflowName: str


def string_to_ansi_color(string: str, color: str) -> str:
    """
    Wrap a string in ANSI color codes.
    """
    colors = {
        "black": "\033[30m",
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "white": "\033[37m",
    }
    return f"{colors[color]}{string}\033[0m"


class ReviewDecision(StrEnum):
    APPROVED = "APPROVED"
    CHANGES_REQUESTED = "CHANGES_REQUESTED"
    REVIEW_REQUIRED = "REVIEW_REQUIRED"
    NONE = ""


review_decision_colors = {
    ReviewDecision.APPROVED: "green",
    ReviewDecision.CHANGES_REQUESTED: "yellow",
    ReviewDecision.REVIEW_REQUIRED: "red",
    ReviewDecision.NONE: "black",
}


def review_decision_to_ansi(review_decision: ReviewDecision) -> str:
    """
    Convert a review decision to an ANSI color escaped string.
    """
    if (
        review_decision not in review_decision_colors
        or review_decision == ReviewDecision.NONE
    ):
        return review_decision.value
    color = review_decision_colors[review_decision]
    return string_to_ansi_color(review_decision.value, color)


class MergeableStatus(str, Enum):
    MERGEABLE = "MERGEABLE"
    CONFLICTING = "CONFLICTING"
    UNKNOWN = "UNKNOWN"


def mergeable_status_to_ansi(mergeable_status: MergeableStatus) -> str:
    """
    Convert a mergeable status to an ANSI color escaped string.
    """
    if mergeable_status == MergeableStatus.MERGEABLE:
        return string_to_ansi_color(mergeable_status.value, "green")
    elif mergeable_status == MergeableStatus.CONFLICTING:
        return string_to_ansi_color(mergeable_status.value, "red")
    return mergeable_status.value


@dataclass
class PrStatus:
    branchname: str
    pr_id: int
    review_decision: ReviewDecision
    is_draft: bool
    is_mergeable: MergeableStatus
    title: str
    status_checks: Optional[dict[str, StatusCheck]] = None

    @classmethod
    def from_gh_cli_dict(cls, gh_cli_dict: dict):
        return cls(
            branchname=gh_cli_dict["headRefName"],
            pr_id=int(gh_cli_dict["number"]),
            review_decision=ReviewDecision(gh_cli_dict["reviewDecision"]),
            is_draft=gh_cli_dict["isDraft"],
            is_mergeable=MergeableStatus(gh_cli_dict["mergeable"]),
            status_checks={
                status["name"]: StatusCheck(
                    **{
                        k: V
                        for k, V in status.items()
                        if k in StatusCheck.__annotations__
                    }
                )
                for status in gh_cli_dict["statusCheckRollup"]
            }
            if gh_cli_dict["statusCheckRollup"]
            else None,
            title=gh_cli_dict["title"],
        )

    @classmethod
    def from_branchname(cls, branchname: str):
        pr_id = get_pr_id_for_branch(branchname)
        if not pr_id:
            return None
        workflow_statuses = []
        for workflow in ghchain.config.workflows:
            workflow_status = WorkflowStatus.create(workflow, branchname)
            if workflow_status:
                workflow_statuses.append(workflow_status)

        # TODO: add statusCheckRollup
        result = run_command(
            [
                "gh",
                "pr",
                "view",
                pr_id,
                "--json",
                "reviewDecision,isDraft,mergeable,statusCheckRollup,title",
            ],
        )
        result_dict = json.loads(result.stdout)

        return cls(
            branchname=branchname,
            pr_id=int(pr_id),
            review_decision=ReviewDecision(result_dict["reviewDecision"]),
            is_draft=result_dict["isDraft"],
            is_mergeable=MergeableStatus(result_dict["mergeable"]),
            status_checks={
                status["name"]: StatusCheck(
                    **{
                        k: V
                        for k, V in status.items()
                        if k in StatusCheck.__annotations__
                    }
                )
                for status in result_dict["statusCheckRollup"]
            }
            if result_dict["statusCheckRollup"]
            else None,
            title=result_dict["title"],
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "branchname": self.branchname,
            "pr_id": self.pr_id,
            "review_decision": self.review_decision.value,
            "is_draft": self.is_draft,
            "is_mergeable": self.is_mergeable.value,
            "title": self.title,
        }



"""Console script for ghchain."""

import click


@click.group(invoke_without_command=True)
@click.option(
    "-p",
    "--create-pr",
    is_flag=True,
    help="If set to True, a pull request will be opened for every commit.",
)
@click.option(
    "--draft",
    is_flag=True,
    help="Create the pull request as a draft. This flag sets --create-pr to True.",
)
@click.option(
    "--with-tests",
    is_flag=True,
    help="Run the github workflows that are specified in the .ghchain.toml config of the repository.",
)
@click.pass_context
def ghchain_cli(ctx, create_pr, draft, with_tests):
    """
    Create a branch for each commit in the stack that doesn't already have one.
    Optionally, create a PR for each branch and run the github workflows that
    are specified in the .ghchain.toml config of the repository.

    Usage:

    ghchain: will create a branch for each commit in the stack that doesn't already have one
      and push it to the remote.

    ghchain --create-pr: will create a branch for each commit in the stack that doesn't already have one,
      push it to the remote and create a PR for each branch.

    ghchain --draft: will create a branch for each commit in the stack that doesn't already have one,
      push it to the remote and create a draft PR for each branch.

    ghchain --with-tests: will create a branch for each commit in the stack that doesn't already have one,
      push it to the remote and run the github workflows that are specified in the
      .ghchain.toml config of the repository.
    """

    from ghchain.stack import Stack

    if ctx.invoked_subcommand is None:
        if draft:
            # when draft is passed, we know that the user wants to publish the PRs
            create_pr = True
        stack = Stack.create()
        for commit in stack.commits:
            stack.process_commit(commit, create_pr, draft, with_tests)

    elif ctx.invoked_subcommand == "help":
        click.echo(ctx.get_help())


@ghchain_cli.command()
@click.argument("commit_sha", type=str)
@click.option(
    "-p",
    "--create-pr",
    is_flag=True,
    help="If set to True, a pull request will be opened for the commit.",
)
@click.option(
    "--draft",
    is_flag=True,
    help="Create the pull request as a draft. This flag sets --create-pr to True.",
)
@click.option(
    "--with-tests",
    is_flag=True,
    help="Run the github workflows that are specified in the .ghchain.toml config of the repository.",
)
def process_commit(commit_sha, create_pr, draft, with_tests):
    """
    Process a single commit by its SHA.
    """
    from ghchain.stack import Stack

    stack = Stack.create()
    commit = next((c for c in stack.commits if c.sha == commit_sha), None)
    if draft:
        # when draft is passed, we know that the user wants to publish the PRs
        create_pr = True
    if commit:
        stack.process_commit(commit, create_pr, draft, with_tests)
    else:
        click.echo(f"Commit with SHA {commit_sha} not found.")


@ghchain_cli.command()
def refresh():
    """
    Update the commit notes with the PR/ workflow statuses for the commits in the stack.
    """
    from ghchain.stack import Stack

    Stack.create()


@ghchain_cli.command()
@click.option(
    "--branch",
    "-b",
    type=str,
    default=None,
    help=("The target branch to which the configured base branch will be updated."),
)
def land(branch):
    """
    Merge the specified branch into the configured base branch.
    """
    from ghchain.git_utils import get_current_branch
    from ghchain.handlers import handle_land

    branch_name = get_current_branch() if branch == "." or not branch else branch
    handle_land(branch_name)


@ghchain_cli.command()
@click.argument(
    "target",
    type=str,
)
@click.option(
    "--interactive",
    "-i",
    is_flag=True,
    help=(
        "Do an interactive rebase. This will allow you to edit "
        "the commit messages and the order of the commits between branches."
    ),
)
def rebase(target, interactive):
    """
    Rebase the current branch onto branch, using 'update-refs' and push every updated branch to the remote.
    """
    from ghchain.git_utils import rebase_onto_target

    rebase_onto_target(target, interactive=interactive)


@ghchain_cli.command()
@click.option(
    "--branch",
    "-b",
    type=str,
    default=None,
    help=("Branch name or PR ID to run the github workflows for."),
)
def run_workflows(branch):
    """
    Run the github workflows that are specified in the .ghchain.toml config of the repository for the specified branch.
    If '.' or nothing is passed, the current branch will be used.
    """
    import ghchain
    from ghchain.git_utils import get_current_branch
    from ghchain.github_utils import (
        get_branch_name_for_pr_id,
    )
    from ghchain.pull_request import PR, get_open_prs, run_tests_on_branch

    branch_name = (
        get_branch_name_for_pr_id(int(branch))
        if branch.isdigit()
        else get_current_branch()
        if branch == "." or not branch
        else branch
    )

    if not branch_name:
        ghchain.logger.error(f"Branch {branch} not found.")
        return

    branches_to_pr: dict[str, PR] = {pr.head_branch: pr for pr in get_open_prs()}

    pull_request = branches_to_pr.get(branch_name)
    if not pull_request:
        ghchain.logger.warning(
            f"Branch {branch_name} not found or does not have a pull request associated with it."
            "Running the test without updating the PR description."
        )

    run_tests_on_branch(branch=branch_name, pull_request=pull_request)


@ghchain_cli.command()
def fix_refs():
    """
    If you messed up your stack with a rebase and lost the connections between the commits and the branches,
    this command will attempt to fix it.
    If you notice the mistake soon enough, you can switch back to your previous ref using reflog... If not,
    this function might help.
    """
    import subprocess

    from ghchain.git_utils import get_commit_message_to_branch_mapping, run_command
    from ghchain.stack import Stack

    stack = Stack.create()

    commit_msg_to_branch = get_commit_message_to_branch_mapping()
    for commit in stack.commits:
        if not commit.branch:
            branch = commit_msg_to_branch.get(commit.message)
            if branch:
                # Ask user if they want to reset the branch to the commit
                if click.confirm(
                    f"Do you want to reset the branch '{branch}' to the commit '{commit.sha}'?"
                    f"\n Commit message:\n{commit.message}"
                ):
                    try:
                        # Reset the branch to the commit
                        run_command(
                            ["git", "branch", "-f", branch, commit.sha], check=True
                        )
                        click.echo(
                            f"Branch '{branch}' has been reset to commit '{commit.sha}'."
                        )
                    except subprocess.CalledProcessError as e:
                        click.echo(
                            f"Failed to reset branch '{branch}' to commit '{commit.sha}': {e}"
                        )
                else:
                    click.echo(
                        f"Skipped resetting branch '{branch}' to commit '{commit.sha}'."
                    )

    click.echo("Finished attempting to fix refs.")


@ghchain_cli.command()
def publish():
    """
    Publish all updated branches in the stack to the remote.
    """
    from ghchain.stack import Stack

    stack = Stack.create()
    stack.publish()


if __name__ == "__main__":
    ghchain_cli()



from unittest.mock import patch

import pytest

from ghchain.commit import Commit


@patch("ghchain.commit.get_issue_url")
@patch.object(Commit, "update_notes")
def test_extract_issue_url(mock_update_notes, mock_get_issue_url):
    # Mock the get_issue_url function to return a test URL
    mock_get_issue_url.return_value = "https://github.com/your_repo/issues/123"

    # Test cases
    test_cases = [
        {
            "message": "[tag] Fix bug in code (#123)",
            "expected_issue_id": 123,
        },
        {
            "message": "Implement new feature (#456)",
            "expected_issue_id": 456,
        },
        {"message": "Refactor code without issue id", "expected_issue_id": None},
    ]

    for case in test_cases:
        commit = Commit(sha="dummy_sha", message=case["message"])
        issue_id = commit.extract_issue_id()
        assert issue_id == case["expected_issue_id"]


if __name__ == "__main__":
    pytest.main()



import os
import subprocess
from pathlib import Path

import pytest
from git import Repo
from loguru import logger

from ghchain.stack import (
    Stack,
    get_current_branch,
)


@pytest.fixture
def temp_git_repo(monkeypatch, tmp_path):
    origin_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        logger.info(f"Setting up temporary directory: {tmp_path}")

        # Initialize a git repository
        subprocess.run(["git", "init"], cwd=tmp_path, check=True)

        # Configure user name and email for the temporary repo
        subprocess.run(
            ["git", "config", "user.name", "Test User"], cwd=tmp_path, check=True
        )
        subprocess.run(
            ["git", "config", "user.email", "test@example.com"],
            cwd=tmp_path,
            check=True,
        )

        # Create initial commit
        initial_file = Path(tmp_path) / "initial_file.txt"
        initial_file.write_text("Initial content")
        subprocess.run(["git", "add", "initial_file.txt"], cwd=tmp_path, check=True)
        subprocess.run(
            ["git", "commit", "-m", "Initial commit"], cwd=tmp_path, check=True
        )

        # Monkeypatch the repo object
        repo = Repo(tmp_path)
        monkeypatch.setattr("ghchain.repo", repo)

        yield tmp_path

    finally:
        os.chdir(origin_cwd)


def run_git_command(command, cwd):
    result = subprocess.run(
        command, cwd=cwd, check=True, capture_output=True, text=True
    )
    return result.stdout.strip()


def test_get_current_branch(temp_git_repo):
    branch = get_current_branch()
    assert branch == "master"


def test_stack_create(temp_git_repo):
    # Create a new branch and commit
    run_git_command(["git", "checkout", "-b", "feature"], temp_git_repo)
    new_file = Path(temp_git_repo) / "new_file.txt"
    new_file.write_text("New content")
    run_git_command(["git", "add", "new_file.txt"], temp_git_repo)
    run_git_command(["git", "commit", "-m", "Add new file"], temp_git_repo)

    stack = Stack.create(base_branch="master")
    assert stack.dev_branch == "feature"
    assert stack.base_branch == "master"
    assert len(stack.commits) == 1
    assert stack.commits[0].sha == run_git_command(
        ["git", "rev-parse", "HEAD"], temp_git_repo
    )
    assert stack.commits[0].message == "Add new file"


if __name__ == "__main__":
    pytest.main(["-v", __file__, "-s", "-k test_create_branches_for_commits"])



import json
import re
import shutil
from typing import Optional


import ghchain
from ghchain.utils import run_command

STACK_LIST_START_MARKER = "<!-- STACK_LIST_START -->"
STACK_LIST_END_MARKER = "<!-- STACK_LIST_END -->"
WORKFLOW_BADGES_START_MARKER = "<!-- WORKFLOW_BADGES_START -->"
WORKFLOW_BADGES_END_MARKER = "<!-- WORKFLOW_BADGES_END -->"


# Verify that gh is installed
if shutil.which("gh") is None:
    raise Exception(
        "gh is not installed. Please install it from https://cli.github.com/"
    )


def run_workflows(workflow_ids: list[str], branch: str) -> list[str]:
    ghchain.logger.info(f"Running workflows for branch {branch}...")
    md_badges = []
    for workflow_id in workflow_ids:
        command = ["gh", "workflow", "run", f"{workflow_id}.yml", "--ref", branch]
        result = run_command(
            command,
            check=True,
        )
        ghchain.logger.debug(result.stdout)
        repo_url = get_repo_url()

        workflow_overview_url = (
            f"{repo_url}/actions/workflows/{workflow_id}.yml?query=branch%3A{branch}"
        )

        md_badges.append(
            (
                f"[![{workflow_id}]({repo_url}/actions/workflows/{workflow_id}.yml/badge.svg"
                f"?branch={branch})]({workflow_overview_url})"
            )
        )
    return md_badges


def get_repo_url() -> str:
    """
    Get the https url of the repository
    """
    repo_url_result = run_command(
        ["gh", "repo", "view", "--json", "url"],
    )
    # weird hack because I didn't get the --jq '.url' to work
    urls = [e for e in repo_url_result.stdout.strip().split('"') if "https" in e]
    if not urls:
        ghchain.logger.error("Failed to get the repository url.")
        raise Exception("Failed to get the repository url.")
    return urls[0]


def get_pr_body(pr_number):
    result = run_command(
        ["gh", "pr", "view", pr_number, "--json", "body"],
    )
    return json.loads(result.stdout)["body"] if result.stdout else ""


def get_workflow_pr_string(md_badges: list[str]):
    workflow_header = "# Workflow Results"
    return "\n".join(
        [
            WORKFLOW_BADGES_START_MARKER,
            workflow_header,
            *md_badges,
            WORKFLOW_BADGES_END_MARKER,
        ]
    )


def get_branch_name_for_pr_id(pr_id) -> Optional[str]:
    result = run_command(
        ["gh", "pr", "list", "--json", "headRefName,number", "--state", "open"],
    )
    prs = json.loads(result.stdout)
    return next((pr["headRefName"] for pr in prs if pr["number"] == pr_id), None)


def get_pr_url_for_branch(branch_name) -> Optional[str]:
    result = run_command(
        ["gh", "pr", "list", "--json", "url,headRefName", "--state", "open"],
    )
    prs = json.loads(result.stdout)
    return next((pr["url"] for pr in prs if pr["headRefName"] == branch_name), None)


def get_pr_url_for_id(pr_id) -> Optional[str]:
    result = run_command(
        ["gh", "pr", "list", "--json", "url,number", "--state", "open"],
    )
    prs = json.loads(result.stdout)
    return next((pr["url"] for pr in prs if pr["number"] == pr_id), None)


def get_latest_id(which: str) -> int:
    result = run_command(
        ["gh", which, "list", "--json", "number", "--state", "all"],
    )
    prs = json.loads(result.stdout)

    sorted_prs = sorted(prs, key=lambda pr: pr["number"], reverse=True)

    if not sorted_prs:
        return -1

    return sorted_prs[0]["number"]


def get_next_gh_id() -> int:
    return max(get_latest_id("pr"), get_latest_id("issue")) + 1


def get_pr_id_for_branch(branch_name) -> Optional[str]:
    pr_url = get_pr_url_for_branch(branch_name)
    if pr_url:
        return pr_url.split("/")[-1]
    return None


def get_pr_approval(pr_id) -> bool:
    result = run_command(
        ["gh", "pr", "view", pr_id, "--json", "reviewDecision"],
    )
    return json.loads(result.stdout)["reviewDecision"] == "APPROVED"


def get_pr_isdraft(pr_id) -> bool:
    result = run_command(
        ["gh", "pr", "view", pr_id, "--json", "isDraft"],
    )
    return json.loads(result.stdout)["isDraft"]


def create_branch_from_issue(issue_id: int, base_commit: str) -> str:
    """
    Create a branch from a GitHub issue and return the branch name.
    """
    # Run the command to create the branch
    result = run_command(["gh", "issue", "develop", str(issue_id)])

    # Check if the command executed successfully
    if result.returncode != 0:
        raise RuntimeError(
            f"Failed to create branch from issue {issue_id}: {result.stderr}"
        )

    # Extract the branch name from the output
    pattern = r"github.com/.*/tree/([^\s]+)"
    match = re.search(pattern, result.stdout)

    if match:
        branch_name = match.group(1)
        # fetch the origin
        ghchain.repo.remotes.origin.fetch()
        # Create a new branch pointing to the base commit
        branch = ghchain.repo.create_head(branch_name, base_commit)

        # Set the branch to track the remote branch
        branch.set_tracking_branch(ghchain.repo.remotes.origin.refs[branch_name])

        # Push the updated branch to the remote
        ghchain.repo.git.push("origin", f"{branch_name}:{branch_name}", force=True)

        return branch_name
    else:
        raise ValueError("Branch name not found in the command output.")



from unittest import mock

import pytest
from ghchain.stack import (
    Commit,
    Stack,
    get_current_branch,
)


def test_get_current_branch(mock_repo):
    mock_repo.active_branch.name = "feature_branch"
    branch = get_current_branch()
    assert branch == "feature_branch"


def test_stack_create(mock_repo):
    mock_repo.git.branch.return_value = "* master\n  develop"
    mock_repo.git.log.return_value = "sha1 commit message\nsha2 another commit message"
    mock_repo.active_branch.name = "feature_branch"

    stack = Stack.create(base_branch="main")
    assert stack.dev_branch == "feature_branch"
    assert stack.base_branch == "main"
    assert len(stack.commits) == 2
    assert stack.commits[0].sha == "sha1"
    assert stack.commits[0].message == "commit message"


@mock.patch("ghchain.git_utils.create_branch_name")
@mock.patch("ghchain.github_utils.get_next_gh_id")
@mock.patch("ghchain.git_utils.git_push")
@mock.patch("ghchain.github_utils.create_pull_request")
@mock.patch("ghchain.github_utils.update_pr_descriptions")
@mock.patch("ghchain.github_utils.run_tests_on_pr")
def test_process_commit(
    mock_run_tests,
    mock_update_pr_descriptions,
    mock_create_pr,
    mock_git_push,
    mock_get_next_gh_id,
    mock_create_branch_name,
):
    # Setup
    mock_get_next_gh_id.return_value = 42
    mock_create_branch_name.return_value = "branch-42"

    commit1 = Commit(sha="sha1", message="commit message 1", branches=set())
    commit2 = Commit(sha="sha2", message="commit message 2", branches=set())
    stack = Stack(commits=[commit1, commit2], dev_branch="feature", base_branch="main")

    # Test process_commit without create_pr
    pr_created = stack.process_commit(commit1, create_pr=False)
    assert not pr_created
    assert "branch-42" in commit1.branches
    assert "branch-42" in commit2.branches

    # Test process_commit with create_pr
    mock_create_pr.return_value = "https://github.com/user/repo/pull/1"
    pr_created = stack.process_commit(
        commit2, create_pr=True, draft=True, with_tests=True
    )
    assert pr_created
    assert "branch-42" in commit2.branches
    assert commit2.pr_url == "https://github.com/user/repo/pull/1"
    mock_run_tests.assert_called_with(
        "branch-42", "https://github.com/user/repo/pull/1"
    )


if "__main__" == __name__:
    pytest.main(["-v", __file__, "-s", "-k test_process_commit", "-x"])



import os
import subprocess

import click

import ghchain
from ghchain.utils import logger, run_command


def git_push(branch_name: str):
    try:
        ghchain.repo.git.push("origin", branch_name)
    except Exception as e:
        logger.error(f"Error pushing branch {branch_name}: {e}")
        raise


def rebase_onto_target(target: str, interactive: bool = False) -> None:
    """
    Rebase the current branch onto a specified target branch or commit hash.

    This function rebases the current branch onto the target specified by the `target` parameter.
    If the `interactive` parameter is set to True, the rebase will be performed interactively.

    Args:
        target (str): The target branch or commit hash to rebase onto.
        interactive (bool, optional): Flag to perform an interactive rebase. Defaults to False.

    The function executes the rebase command using subprocess. If the rebase is interactive, it constructs
    the command accordingly and executes it with `shell=True`. In case of non-interactive rebase, it calls
    `run_command` with the constructed command.

    During the rebase, if conflicts are detected (indicated by "CONFLICT" or "needs merge" in the command output),
    the user is prompted to resolve them. After resolving conflicts, pressing Enter continues the rebase process.

    After a successful rebase, the function extracts the names of branches that were rebased (if any are reported
    in the stderr output) and pushes them to the remote repository using `--force-with-lease` to ensure safety.

    Note:
        - The `GIT_EDITOR` environment variable is set to "true" to prevent the editor from opening during
          `git rebase --continue`

    """
    rebase_command = ["git", "rebase", "--update-refs", target]
    if interactive:
        rebase_command.insert(2, "--interactive")
        rebase_command = " ".join(rebase_command)
        result = subprocess.run(
            rebase_command, shell=True, stderr=subprocess.PIPE, text=True
        )

    else:
        # don't check because otherwise it raises an error when there are conflicts
        result = run_command(rebase_command, shell=interactive, check=False)

    click.echo(f"Rebase command output: {result.stdout} {result.stderr}")

    if result.stdout:
        while any(e in result.stdout for e in ["CONFLICT", "needs merge"]):
            click.echo(
                "Conflicts detected during rebase. Please resolve them and then press Enter to continue."
            )
            input()
            result = run_command(
                ["git", "rebase", "--continue"],
                env={
                    **os.environ,
                    "GIT_EDITOR": "true",
                },  # to prevent the editor from opening
                check=False,
            )
            print(f"Rebase command output: {result.stdout}\n{result.stderr}")

    branches = [
        line.strip().replace("refs/heads/", "")
        for line in result.stderr.split("\n")
        if line.strip() and "refs/heads/" in line and "Successfully rebased" not in line
    ]

    # push each branch to origin
    logger.info(f"Pushing branches: {branches}")
    command = ["git", "push", "--force-with-lease", "origin"] + branches
    run_command(command, check=True)


def create_branch_name(branch_name_template: str, next_pr_id: int):
    # Get the git author name
    author_name = (
        ghchain.repo.config_reader().get_value("user", "name").replace(" ", "_").lower()
    )

    branch_name = branch_name_template.format(
        git_config_author=author_name, pr_id=next_pr_id
    )

    return branch_name


def get_current_branch() -> str:
    """
    Get the current branch name.
    """
    try:
        return ghchain.repo.active_branch.name
    except TypeError as e:
        ghchain.logger.error(f"Error getting current branch: {e}")
        raise


def get_all_branches(remote: bool = False) -> list[str]:
    """
    Get all branches in the repository.
    """
    if remote:
        return get_all_remote_branches()
    return [branch.name for branch in ghchain.repo.branches]


def get_all_remote_branches() -> list[str]:
    """
    Get all remote branches in the repository.
    """
    remote_branches = []
    for remote in ghchain.repo.remotes:
        for branch in remote.refs:
            if "HEAD" not in branch.name:
                remote_branches.append(branch.name.replace(f"{remote.name}/", ""))
    return remote_branches


def get_commit_message_to_branch_mapping() -> dict[str, str]:
    """
    Get a mapping of commit messages to branch names.

    This function returns a dictionary where the keys are commit messages and the values are the branch names
    that contain the corresponding commit.

    Returns:
        dict[str, str]: A dictionary mapping commit messages to branch names.
    """
    return {
        branch.commit.message.replace("\n", ""): branch.name
        for branch in ghchain.repo.branches
    }


def get_issue_url(issue_id: int) -> str:
    """
    Get the GitHub issue URL using the issue ID and the remote repository URL.
    """
    remote_url = ghchain.repo.remotes.origin.url

    # Extract the owner and repo name from the remote URL
    splits = remote_url.split("/")
    owner, repo_name = splits[-2].split(":")[-1], splits[-1].replace(".git", "")

    # Construct the issue URL
    issue_url = f"https://github.com/{owner}/{repo_name}/issues/{issue_id}"

    return issue_url



import json
import re
from dataclasses import dataclass
from typing import Optional

import click

import ghchain
from ghchain.github_utils import (
    STACK_LIST_END_MARKER,
    STACK_LIST_START_MARKER,
    WORKFLOW_BADGES_END_MARKER,
    WORKFLOW_BADGES_START_MARKER,
    get_workflow_pr_string,
    run_workflows,
)
from ghchain.status import PrStatus
from ghchain.utils import run_command


@dataclass
class PR:
    """
    Class representing a github pull request.
    """

    pr_id: int
    pr_url: str
    pr_status: PrStatus | None
    head_branch: str
    body: str
    title: str
    commits: list[str]

    @classmethod
    def create_pull_request(
        cls, base_branch, head_branch, title, body, commit_sha, draft=False
    ) -> Optional["PR"]:
        ghchain.logger.info(
            f"Creating pull request from {head_branch} to {base_branch}."
        )

        command = [
            "gh",
            "pr",
            "create",
            "--base",
            base_branch,
            "--head",
            head_branch,
            "--title",
            title,
            "--body",
            body,
        ] + (["--draft"] if draft else [])
        result = run_command(command)
        url_match = re.search(r"https://github\.com/.+/pull/\d+", result.stdout)
        if url_match:
            click.echo(
                f"Pull request created: {url_match.group(0)} (Draft: {'Yes' if draft else 'No'})"
            )
            pr_url = url_match.group(0)
            pr_id = int(pr_url.rstrip("/").split("/")[-1])
            return cls(
                pr_url=url_match.group(0),
                pr_id=pr_id,
                pr_status=None,
                head_branch=head_branch,
                body=body,
                title=title,
                commits=[commit_sha],
            )

        click.echo("Failed to create pull request.")
        return None


def get_open_prs() -> list[PR]:
    gh_command = [
        "gh",
        "pr",
        "list",
        "--json",
        "url,headRefName,commits,number,isDraft,latestReviews,reviewDecision,mergeable,statusCheckRollup,title,body",
        "--state",
        "open",
    ]
    result = run_command(gh_command, check=True)
    result_dicts = json.loads(result.stdout)
    return [
        PR(
            pr_status=PrStatus.from_gh_cli_dict(result_dict),
            pr_id=int(result_dict["number"]),
            pr_url=result_dict["url"],
            head_branch="headRefName",
            body=result_dict["body"],
            title=result_dict["title"],
            commits=[commit["oid"] for commit in result_dict["commits"]],
        )
        for result_dict in result_dicts
    ]


def update_pr_stacklist_description(current_body, pr_url, pr_stack: list[PR]) -> str:
    """Update all PRs in the stack with the full stack list in their descriptions."""
    stack_list_lines = []

    for pr in pr_stack:
        # Highlight the current PR with an arrow
        if pr.pr_url == pr_url:
            stack_list_lines.insert(0, f"- -> {pr.pr_url}")
        else:
            stack_list_lines.insert(0, f"- {pr.pr_url}")

    stack_list_md = (
        (
            f"{STACK_LIST_START_MARKER}\nStack from [ghchain]"
            "(https://github.com/Jimmy2027/ghchain) (oldest at the bottom):\n"
        )
        + "\n".join(stack_list_lines)
        + f"\n{STACK_LIST_END_MARKER}"
    )

    if (
        STACK_LIST_START_MARKER in current_body
        and STACK_LIST_END_MARKER in current_body
    ):
        updated_body = re.sub(
            f"{STACK_LIST_START_MARKER}.*?{STACK_LIST_END_MARKER}",
            stack_list_md,
            current_body,
            flags=re.DOTALL,
        )
    else:
        updated_body = f"{current_body}\n\n{stack_list_md}"
    return updated_body


def update_pr_descriptions(pr_stack: list[PR]):
    """
    Update all PRs in the stack with the full stack list in their descriptions, highlighting the current PR.
    run_tests is a tuple of the pr_url and the branch name.
    """
    prs_str = f"{', '.join([f'#{pr.pr_id}' for pr in pr_stack])}"
    ghchain.logger.info(f"Updating PR descriptions of {prs_str}.")

    # TODO: check that I'm the author of the PR
    for pr in pr_stack:
        updated_body = update_pr_stacklist_description(pr.body, pr.pr_url, pr_stack)

        pr.body = updated_body

        run_command(
            ["gh", "pr", "edit", str(pr.pr_id), "--body", updated_body],
            check=True,
        )

        click.echo(f"PR description updated for PR #{pr.pr_id}.")


def run_tests_on_branch(branch: str, pull_request: PR | None = None):
    """
    Runs the configured workflows on the specified branch
    and updates the PR description with the workflow results if a PR URL is provided.

    Args:
        branch (str): The branch on which to run the workflows.
        pull_request: (str | None, optional): The pull request object of the pull request to update.
            If None, the PR description is not updated.

    Returns:
        None
    """
    if not ghchain.config.workflows:
        ghchain.logger.error("No workflows found in the config.")
        return

    md_badges = run_workflows(ghchain.config.workflows, branch)
    if pull_request is None:
        ghchain.logger.debug(
            f"No PR found for branch {branch}. Not updating PR description."
        )
        return
    workflow_string = get_workflow_pr_string(md_badges)

    current_body = pull_request.body

    if (
        WORKFLOW_BADGES_START_MARKER in current_body
        and WORKFLOW_BADGES_END_MARKER in current_body
    ):
        updated_body = re.sub(
            f"{WORKFLOW_BADGES_START_MARKER}.*?{WORKFLOW_BADGES_END_MARKER}",
            workflow_string,
            current_body,
            flags=re.DOTALL,
        )
    else:
        updated_body = f"{current_body}\n{workflow_string}"

    pull_request.body = updated_body

    run_command(
        [
            "gh",
            "pr",
            "edit",
            pull_request.pr_url.split("/")[-1],
            "--body",
            updated_body,
        ],
        check=True,
    )
    ghchain.logger.debug(f"PR description updated for PR {pull_request.pr_url}.")
